/* Copyright (C) 2011, 2012, 2013 by Bundesanstalt für Gewässerkunde
 * Software engineering by Intevation GmbH
 *
 * This file is Free Software under the GNU AGPL (>=v3)
 * and comes with ABSOLUTELY NO WARRANTY! Check out the
 * documentation coming with Dive4Elements River for details.
 */

package org.dive4elements.river.exports;

import java.io.OutputStream;
import java.text.DateFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.log4j.Logger;
import org.dive4elements.artifacts.CallMeta;
import org.dive4elements.artifacts.common.utils.Config;
import org.dive4elements.river.artifacts.D4EArtifact;
import org.dive4elements.river.artifacts.access.RangeAccess;
import org.dive4elements.river.artifacts.model.CalculationResult;
import org.dive4elements.river.artifacts.model.WKmsJRDataSource;
import org.dive4elements.river.artifacts.model.WQ;
import org.dive4elements.river.artifacts.model.WQKms;
import org.dive4elements.river.artifacts.resources.Resources;
import org.dive4elements.river.model.Gauge;
import org.dive4elements.river.model.River;
import org.dive4elements.river.utils.Formatter;
import org.dive4elements.river.utils.RiverUtils;

import au.com.bytecode.opencsv.CSVWriter;
import net.sf.jasperreports.engine.JRException;
import net.sf.jasperreports.engine.JasperExportManager;
import net.sf.jasperreports.engine.JasperFillManager;
import net.sf.jasperreports.engine.JasperPrint;

/**
 * @author <a href="mailto:ingo.weinzierl@intevation.de">Ingo Weinzierl</a>
 */
public class ComputedDischargeCurveExporter extends AbstractExporter {

    /** The log used in this exporter. */
    private static Logger log = Logger.getLogger(ComputedDischargeCurveExporter.class);

    public static final String CSV_W_HEADER = "export.computed.discharge.curve.csv.header.w";

    public static final String CSV_Q_HEADER = "common.export.csv.header.q";

    public static final String DEFAULT_CSV_W_HEADER = "W [NN + m]";
    public static final String DEFAULT_CSV_Q_HEADER = "Q [m\u00b3/s]";

    public static final String PDF_HEADER_MODE = "export.computed.discharge.pdf.mode";
    public static final String PDF_HEADER_CALC_MODE = "export.computed.discharge.pdf.calc.mode";
    public static final String JASPER_FILE = "export.computed.discharge.pdf.file";

    protected List<WQKms> data;

    protected String wUnit;
    protected String riverUnit;
    protected String gaugeName;
    protected double gaugeDatum;
    protected Date validSince;

    public ComputedDischargeCurveExporter() {
        this.data = new ArrayList<>();
    }

    @Override
    protected void addData(Object d) {
        if (d instanceof CalculationResult) {
            d = ((CalculationResult) d).getData();
        }
        WQKms referenceWQ = null; // used for gauge / unit observations
        if (d instanceof WQKms[]) {
            this.data.addAll(Arrays.asList((WQKms[]) d));
            // If there is a unit mix in this list
            // we are screwed anyway.
            referenceWQ = ((WQKms[]) d)[0];
        } else if (d instanceof WQKms) {
            this.data.add((WQKms) d);
            referenceWQ = (WQKms) d;
        } else {
            log.warn("Can't add data for export. Unkown data type " + d.getClass().getName());
            return;
        }
        if (referenceWQ != null) {
            final D4EArtifact arti = (D4EArtifact) this.master;
            final River river = RiverUtils.getRiver(arti);
            this.riverUnit = river.getWstUnit().getName();
            final RangeAccess rangeAccess = new RangeAccess(arti);

            final double[] kms = rangeAccess.getKmRange();

            final Gauge gauge = river.determineGaugeAtStation(kms[0]);
            if (gauge != null) {
                this.wUnit = "cm";
                this.gaugeName = gauge.getName();
                this.gaugeDatum = gauge.getDatum().doubleValue();

                // Now convert the data to cm because we are at gauge
                final List<WQKms> newData = new ArrayList<>();
                for (final WQKms d2 : this.data) {
                    newData.add(new WQKms(d2.getKms(), WQ.getFixedWQforExportAtGauge(d2, gauge.getDatum())));
                }
                this.data = newData; // All hail the garbage collector

                this.validSince = gauge.fetchMasterDischargeTable().getTimeInterval().getStartTime();
            } else {
                this.gaugeName = "";
                this.validSince = null;
                this.gaugeDatum = Double.NaN;
            }
        }
    }

    @Override
    protected void writeCSVData(final CSVWriter writer) {
        log.info("ComputedDischargeCurveExporter.writeData");

        writeCSVHeader(writer);

        final NumberFormat wf = getWFormatter();
        final NumberFormat qf = getQFormatter();

        double[] res = new double[3];

        for (final WQKms wqkms : this.data) {
            final int size = wqkms.size();

            for (int i = 0; i < size; i++) {
                res = wqkms.get(i, res);

                writer.writeNext(new String[] { wf.format(res[0]), qf.format(res[1]) });
            }
        }
    }

    protected void writeCSVHeader(final CSVWriter writer) {
        log.debug("ComputedDischargeCurveExporter.writeCSVHeader");

        final RangeAccess access = new RangeAccess((D4EArtifact) this.master);
        final double[] km = access.getLocations();
        // If we are not at gauge (cm) use the river unit
        final String realUnit = "cm".equals(this.wUnit) ? "cm" : this.riverUnit;
        final String header = msg(CSV_W_HEADER, DEFAULT_CSV_W_HEADER, new Object[] { realUnit });

        writer.writeNext(new String[] { header, msg(CSV_Q_HEADER, DEFAULT_CSV_Q_HEADER) });
    }

    /**
     * Returns the number formatter for W values.
     *
     * @return the number formatter for W values.
     */
    @Override
    protected NumberFormat getWFormatter() {
        if ("cm".equals(this.wUnit)) {
            return Formatter.getFormatter(this.context, 0, 0);
        }
        return Formatter.getComputedDischargeW(this.context);
    }

    /**
     * Returns the number formatter for Q values.
     *
     * @return the number formatter for Q values.
     */
    @Override
    protected NumberFormat getQFormatter() {
        return Formatter.getComputedDischargeQ(this.context);
    }

    @Override
    protected void writePDF(final OutputStream out) {
        final WKmsJRDataSource source = createJRData();

        final String jasperFile = Resources.getMsg(this.context.getMeta(), JASPER_FILE, "/jasper/computed-discharge_en.jasper");
        final String confPath = Config.getConfigDirectory().toString();

        final Map parameters = new HashMap();
        parameters.put("ReportTitle", "Exported Data");
        try {
            final JasperPrint print = JasperFillManager.fillReport(confPath + jasperFile, parameters, source);
            JasperExportManager.exportReportToPdfStream(print, out);
        }
        catch (final JRException je) {
            log.warn("Error generating PDF Report!");
            je.printStackTrace();
        }
    }

    protected WKmsJRDataSource createJRData() {
        final WKmsJRDataSource source = new WKmsJRDataSource();

        addMetaData(source);
        addWQData(source);

        return source;
    }

    protected void addMetaData(final WKmsJRDataSource source) {
        final CallMeta meta = this.context.getMeta();

        final D4EArtifact flys = (D4EArtifact) this.master;
        source.addMetaData("gauge", this.gaugeName);
        if (!Double.isNaN(this.gaugeDatum)) {
            final NumberFormat mf = Formatter.getMeterFormat(this.context);
            source.addMetaData("datum", mf.format(this.gaugeDatum) + " " + this.riverUnit);
        } else {
            source.addMetaData("datum", "");
        }

        source.addMetaData("river", RiverUtils.getRivername(flys));

        final Locale locale = Resources.getLocale(meta);
        final DateFormat df = DateFormat.getDateInstance(DateFormat.SHORT, locale);

        source.addMetaData("date", df.format(new Date()));

        source.addMetaData("wUnit", "cm".equals(this.wUnit) ? "cm" : this.riverUnit);

        final RangeAccess rangeAccess = new RangeAccess(flys);
        final double[] kms = rangeAccess.getKmRange();
        source.addMetaData("range", Formatter.getCalculationKm(this.context.getMeta()).format(kms[0]));

        if (!"cm".equals(this.wUnit)) {
            source.addMetaData("valid_since", "");
            source.addMetaData("calculation", Resources.getMsg(locale, PDF_HEADER_CALC_MODE, "Computed Discharge"));
        } else {
            source.addMetaData("valid_since", this.validSince == null ? "" : df.format(this.validSince));
            source.addMetaData("calculation", Resources.getMsg(locale, PDF_HEADER_MODE, "Discharge"));
        }
    }

    protected void addWQData(final WKmsJRDataSource source) {
        final NumberFormat wf = getWFormatter();
        final NumberFormat qf = getQFormatter();

        double[] res = new double[3];

        for (final WQKms wqkms : this.data) {
            final int size = wqkms.size();

            for (int i = 0; i < size; i++) {
                res = wqkms.get(i, res);

                source.addData(new String[] { "", // Empty, the WKmsJRDtasource stores km here.
                        wf.format(res[0]), qf.format(res[1]) });
            }
        }
    }
}
// vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
